# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.
from jupyter_core.paths import jupyter_data_dir
from whoosh.index import create_in, open_dir, exists_in, LockError
from whoosh.fields import Schema, TEXT, ID, STORED
from whoosh.query import AndMaybe, Term
from whoosh.qparser import MultifieldParser
import nbformat
import io
import os
import json
import shutil

# Use the built-in version of scandir if possible, otherwise
# use the scandir module version
try:
    from os import scandir
except ImportError:
    from scandir import scandir

class Index(object):
    def __init__(self, work_dir):
        self.work_dir = work_dir
        self._init_index()
        
    def _init_index(self, reset=False):
        index_path = os.path.join(jupyter_data_dir(), 'index')
        
        # clear out old index if requested
        if reset:
            shutil.rmtree(index_path, True)
        
        # make sure there's a path to store the index data
        if not os.path.exists(index_path):
            os.makedirs(index_path)
            
        if not exists_in(index_path):
            # create an index with the current schema
            schema = Schema(basename=TEXT(stored=True, field_boost=5.0), 
                            dirname=ID(stored=True),
                            path=ID(stored=True, unique=True), 
                            content=TEXT(stored=False), 
                            time=STORED)
            self.ix = create_in(index_path, schema)
        else:
            # open the existing index
            self.ix = open_dir(index_path)
            
        # build a query parser based on the current schema
        self.query_parser = MultifieldParser(["content", "basename", "dirname"], self.ix.schema)
    
    def _file_to_document(self, filename, m_time):
        content = u''
        
        # get content for notebooks only
        if filename.endswith('.ipynb'):
            with io.open(filename, 'r', encoding='utf-8') as f:
                try:
                    notebook = nbformat.read(f, 4)
                    content = u'\n'.join(cell.get('source', u'') for cell in notebook['cells'])
                except (Exception):
                    pass
        return dict(
            basename=os.path.basename(filename),
            dirname=os.path.dirname(filename),
            path=filename,
            content=content,
            time=m_time
        )
    
    def _scan_disk(self, on_disk, path):
        for entry in scandir(path):
            if not entry.name.startswith('.') and entry.is_dir():
                self._scan_disk(on_disk, entry.path)
            elif entry.is_file():
                on_disk[entry.path] = entry.stat().st_mtime
        return on_disk
                
    def _scan_index(self):
        in_index = {}
        with self.ix.searcher() as searcher:
            for fields in searcher.all_stored_fields():
                in_index[fields['path']] = fields['time']
        return in_index
                
    def _compute_ops(self, on_disk, in_index):
        in_index_set = set(in_index.keys())
        on_disk_set = set(on_disk.keys())

        to_remove = in_index_set - on_disk_set
        to_add = on_disk_set - in_index_set
        to_update = on_disk_set.intersection(in_index_set)
        
        return to_add, to_remove, to_update
        
    def _add_to_index(self, writer, to_add, on_disk):
        for filename in to_add:
            meta = self._file_to_document(filename, on_disk[filename])
            writer.add_document(**meta)
        
    def _remove_from_index(self, writer, to_remove):
        for filename in to_remove:
            writer.delete_by_term('path', filename)
            
    def _update_in_index(self, writer, to_update, on_disk, in_index):
        for filename in to_update:
            # only update if modification time differs
            if on_disk[filename] == in_index[filename]:
                continue
            meta = self._file_to_document(filename, on_disk[filename])
            writer.update_document(**meta)
    
    def update_index(self):
        '''
        Updates the index based on the disk/index delta.
        '''
        on_disk = {}
        on_disk = self._scan_disk(on_disk, self.work_dir)
        in_index = self._scan_index()
        to_add, to_remove, to_update = self._compute_ops(on_disk, in_index)
        
        #print(len(to_add), len(to_remove), len(to_update)) 
        
        try:
            writer = self.ix.writer()
        except LockError:
            # skip index updates: locked by another process
            pass
        else:
            self._add_to_index(writer, to_add, on_disk)
            self._remove_from_index(writer, to_remove)
            self._update_in_index(writer, to_update, on_disk, in_index)
            writer.commit()

    def search(self, query_string, limit=25, cwd=os.getcwd()):
        '''
        Searches the index given a query string.
        '''
        with self.ix.searcher() as searcher:
            # parse user query
            query = self.query_parser.parse(query_string)
            # improve the score of files in the same directory
            query = AndMaybe(query, Term('dirname', cwd))
            results = searcher.search(query, limit=limit)
            # return dict copies: results not valid outside of context
            return ([dict(
                basename=result['basename'],
                dirname=result['dirname'],
                path=result['path']
            ) for result in results], len(results))

    def reset_index(self):
        '''
        Clears the search index.
        '''
        self._init_index(True)



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.
from .inject import InjectMagic
from . import search
from . import uploads
from . import loader
from jupyter_core.paths import jupyter_runtime_dir
import os
import json


def _jupyter_server_extension_paths():
    '''API for server extension installation on notebook 4.2'''
    return [{
        "module": "jupyter_cms"
    }]


def _jupyter_nbextension_paths():
    '''API for JS extension installation on notebook 4.2'''
    return [{
        'section': 'notebook',
        'src': 'nbextension',
        'dest': 'jupyter_cms',
        'require': 'jupyter_cms/notebook/main'
    },
    {
        'section': 'tree',
        'src': 'nbextension',
        'dest': 'jupyter_cms',
        'require': 'jupyter_cms/dashboard/main'
    },
    {
        'section': 'edit',
        'src': 'nbextension',
        'dest': 'jupyter_cms',
        'require': 'jupyter_cms/editor/main'
    }]

# Use the built-in version of scandir if possible, otherwise
# use the scandir module version
try:
    from os import scandir
except ImportError:
    from scandir import scandir


def load_ipython_extension(ipython):
    # use the configured working directory if we can find it
    work_dir = None
    for filename in scandir(jupyter_runtime_dir()):
        if filename.name.startswith('nbserver-') and filename.name.endswith('.json'):
            with open(filename.path, 'r') as fh:
                nbserver = json.load(fh)
                work_dir = nbserver['notebook_dir']
                break
    if work_dir is None:
        # fall back on an environment variable or ultimately the pwd
        work_dir = os.getenv('WORK', '.')

    loader.enable(work_dir)
    ipython.push({'load_notebook': loader.load_notebook})
    ipython.register_magics(InjectMagic(ipython))


def unload_ipython_extension(ipython):
    ipython.drop_by_id({'load_notebook': loader.load_notebook})
    loader.disable()
    del ipython.magics_manager.magics['line']['inject']
    del ipython.magics_manager.registry['InjectMagic']


def load_jupyter_server_extension(nb_app):
    '''
    Loads all extensions within this package.
    '''
    nb_app.log.info('Loaded jupyter_cms')
    search.load_jupyter_server_extension(nb_app)
    uploads.load_jupyter_server_extension(nb_app)



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.
from notebook.utils import url_path_join, url_unescape
from notebook.base.handlers import IPythonHandler, path_regex
from tornado import web
import os

class UploadsHandler(IPythonHandler):
    def initialize(self, work_dir):
        self.work_dir = work_dir

    @web.authenticated
    def post(self, path):
        '''
        Write uploaded files to disk.

        :param path:
        '''
        files = self.request.files
        if not len(files):
            raise web.HTTPError(400, 'missing files to upload')
        root_path = os.path.join(self.work_dir, path.strip('/'))
        written_paths = []
        for filename, metas in files.items():
            path = url_unescape(os.path.join(root_path, filename))
            with open(path, 'wb') as fh:
                fh.write(metas[0].body)
            written_paths.append(path)
        self.finish({
            'files' : written_paths,
            'path' : url_unescape(root_path)
        })

def load_jupyter_server_extension(nb_app):
    web_app = nb_app.web_app
    host_pattern = '.*$'
    route_pattern = url_path_join(web_app.settings['base_url'], 
        '/uploads%s' % path_regex)
    handler_kwargs = dict(work_dir=nb_app.notebook_dir)
    web_app.add_handlers(host_pattern, [
        (route_pattern, UploadsHandler, handler_kwargs)
    ])



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.
import errno
import os.path
import sys

from ._version import __version__

from jupyter_core.paths import jupyter_config_dir
from notebook.services.config import ConfigManager
from notebook.nbextensions import (InstallNBExtensionApp, EnableNBExtensionApp,
    DisableNBExtensionApp, flags, aliases)

try:
    from notebook.extensions import BaseExtensionApp
    _new_extensions = True
except ImportError:
    try:
        from notebook.nbextensions import BaseNBExtensionApp
        BaseExtensionApp = BaseNBExtensionApp
        _new_extensions = True
    except ImportError:
        BaseExtensionApp = object
        _new_extensions = False

from traitlets import Unicode
from traitlets.config.application import catch_config_error
from traitlets.config.application import Application

# Make copies to reuse flags and aliases
INSTALL_FLAGS = {}
INSTALL_FLAGS.update(flags)

INSTALL_ALIASES = {}
INSTALL_ALIASES.update(aliases)
del INSTALL_ALIASES['destination']


def makedirs(path):
    '''
    mkdir -p and ignore existence errors compatible with Py2/3.
    '''
    try:
        os.makedirs(path)
    except OSError as e:
        if e.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise


class ExtensionInstallApp(InstallNBExtensionApp):
    '''Subclass that installs this particular extension.'''
    name = u'jupyter-cms-extension-install'
    description = u'Install the jupyter_cms extension'

    flags = INSTALL_FLAGS
    aliases = INSTALL_ALIASES

    examples = """
        jupyter cms install
        jupyter cms install --user
        jupyter cms install --prefix=/path/to/prefix
        jupyter cms install --nbextensions=/path/to/nbextensions
    """

    destination = Unicode('')

    def _classes_default(self):
        return [ExtensionInstallApp, InstallNBExtensionApp]

    def start(self):
        here = os.path.abspath(os.path.join(os.path.dirname(__file__)))

        self.log.info("Installing jupyter_cms JS notebook extensions")
        self.extra_args = [os.path.join(here, 'nbextension')]
        self.destination = 'jupyter_cms'
        self.install_extensions()


class ExtensionActivateApp(EnableNBExtensionApp):
    '''Subclass that activates this particular extension.'''
    name = u'jupyter-cms-extension-activate'
    description = u'Activate the jupyter_cms extension'

    flags = {}
    aliases = {}

    examples = """
        jupyter cms activate
    """

    def _classes_default(self):
        return [ExtensionActivateApp, EnableNBExtensionApp]

    def enable_server_extension(self, extension):
        '''Enables the server side extension in the user config.'''
        server_cm = ConfigManager(config_dir=jupyter_config_dir())

        makedirs(server_cm.config_dir)

        cfg = server_cm.get('jupyter_notebook_config')
        server_extensions = (
            cfg.setdefault('NotebookApp', {})
            .setdefault('server_extensions', [])
        )
        if extension not in server_extensions:
            cfg['NotebookApp']['server_extensions'] += [extension]
        server_cm.update('jupyter_notebook_config', cfg)

    def start(self):
        self.log.info("Activating jupyter_cms notebook server extensions")
        self.enable_server_extension('jupyter_cms')

        self.log.info("Activating jupyter_cms JS notebook extensions")
        self.section = "notebook"
        self.enable_nbextension("jupyter_cms/notebook/main")
        self.section = "tree"
        self.enable_nbextension("jupyter_cms/dashboard/main")
        self.section = "edit"
        self.enable_nbextension("jupyter_cms/editor/main")

        self.log.info("Done. You may need to restart the Jupyter notebook server for changes to take effect.")


class ExtensionDeactivateApp(DisableNBExtensionApp):
    '''Subclass that deactivates this particular extension.'''
    name = u'jupyter-cms-extension-deactivate'
    description = u'Deactivate the jupyter_cms extension'

    flags = {}
    aliases = {}

    examples = """
        jupyter cms deactivate
    """

    def _classes_default(self):
        return [ExtensionDeactivateApp, DisableNBExtensionApp]

    def disable_server_extension(self, extension):
        '''Disables the server side extension in the user config.'''
        server_cm = ConfigManager(config_dir=jupyter_config_dir())

        makedirs(server_cm.config_dir)

        cfg = server_cm.get('jupyter_notebook_config')
        if ('NotebookApp' in cfg and
            'server_extensions' in cfg['NotebookApp'] and
            extension in cfg['NotebookApp']['server_extensions']):
            cfg['NotebookApp']['server_extensions'].remove(extension)

        server_cm.update('jupyter_notebook_config', cfg)

        server_extensions = (
            cfg.setdefault('NotebookApp', {})
            .setdefault('server_extensions', [])
        )
        if extension in server_extensions:
            cfg['NotebookApp']['server_extensions'].remove(extension)
        server_cm.update('jupyter_notebook_config', cfg)

    def start(self):
        self.log.info("Deactivating jupyter_cms notebook server extensions")
        self.disable_server_extension('jupyter_cms')

        self.log.info("Deactivating jupyter_cms JS notebook extensions")
        self.section = "notebook"
        self.disable_nbextension("jupyter_cms/notebook/main")
        self.section = "tree"
        self.disable_nbextension("jupyter_cms/dashboard/main")
        self.section = "edit"
        self.disable_nbextension("jupyter_cms/editor/main")

        self.log.info("Done. You may need to restart the Jupyter notebook server for changes to take effect.")


class ExtensionQuickSetupApp(BaseExtensionApp):
    """Installs and enables all parts of this extension"""
    name = "jupyter cms quick-setup"
    version = __version__
    description = "Installs and enables all features of the jupyter_cms extension"

    def start(self):
        self.argv.extend(['--py', 'jupyter_cms'])

        from notebook import serverextensions
        install = serverextensions.EnableServerExtensionApp()
        install.initialize(self.argv)
        install.start()
        from notebook import nbextensions
        install = nbextensions.InstallNBExtensionApp()
        install.initialize(self.argv)
        install.start()
        enable = nbextensions.EnableNBExtensionApp()
        enable.initialize(self.argv)
        enable.start()


class ExtensionQuickRemovalApp(BaseExtensionApp):
    """Disables and uninstalls all parts of this extension"""
    name = "jupyter cms quick-remove"
    version = __version__
    description = "Disables and removes all features of the jupyter_cms extension"

    def start(self):
        self.argv.extend(['--py', 'jupyter_cms'])

        from notebook import nbextensions
        enable = nbextensions.DisableNBExtensionApp()
        enable.initialize(self.argv)
        enable.start()
        install = nbextensions.UninstallNBExtensionApp()
        install.initialize(self.argv)
        install.start()
        from notebook import serverextensions
        install = serverextensions.DisableServerExtensionApp()
        install.initialize(self.argv)
        install.start()


class ExtensionApp(Application):
    '''CLI for extension management.'''
    name = u'jupyter_cms extension'
    description = u'Utilities for managing the jupyter_cms extension'
    examples = ""

    subcommands = dict()

    if _new_extensions:
        subcommands.update({
            "quick-setup": (
                ExtensionQuickSetupApp,
                "Install and enable everything in the package"
            ),
            "quick-remove": (
                ExtensionQuickRemovalApp,
                "Disable and uninstall everything in the package"
            )
        })
    else:
        subcommands.update(dict(
            install=(
                ExtensionInstallApp,
                "Install the extension"
            ),
            activate=(
                ExtensionActivateApp,
                "Activate the extension"
            ),
            deactivate=(
                ExtensionDeactivateApp,
                "Deactivate the extension"
            ),
        ))

    def _classes_default(self):
        classes = super(ExtensionApp, self)._classes_default()

        # include all the apps that have configurable options
        for appname, (app, help) in self.subcommands.items():
            if len(app.class_traits(config=True)) > 0:
                classes.append(app)

    @catch_config_error
    def initialize(self, argv=None):
        super(ExtensionApp, self).initialize(argv)

    def start(self):
        # check: is there a subapp given?
        if self.subapp is None:
            self.print_help()
            sys.exit(1)

        # This starts subapps
        super(ExtensionApp, self).start()


def main():
    ExtensionApp.launch_instance()



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.
from .index import Index
from notebook.utils import url_path_join
from notebook.base.handlers import IPythonHandler
from tornado import web
import os

class SearchHandler(IPythonHandler):
    def initialize(self, work_dir):
        self.index = Index(work_dir)
        self.work_dir = work_dir
        self.work_dir_len = len(self.work_dir)+1

    @web.authenticated
    def get(self):
        query_string = self.get_query_argument('qs')
        reindex = bool(self.get_query_argument('reindex', 'true') == 'true')

        if reindex:
            self.index.update_index()

        results, total = self.index.search(query_string)

        for result in results:
            rel_path = result['path'][self.work_dir_len:]
            if rel_path.endswith('.ipynb'):
                # take it at face value that the extension implies notebook
                url = url_path_join(self.base_url, 'notebooks', rel_path)
            else:
                url = url_path_join(self.base_url, 'edit', rel_path)
            # Add URLs
            result['url'] = url
            result['tree_url'] = url_path_join(self.base_url, 'tree', os.path.dirname(rel_path))
            # Add relative paths
            result['rel_dirname'] = os.path.dirname(rel_path)
            result['rel_path'] = rel_path
        
        self.write(dict(results=results, total=total))
        self.finish()

def load_jupyter_server_extension(nb_app):
    web_app = nb_app.web_app
    host_pattern = '.*$'
    route_pattern = url_path_join(web_app.settings['base_url'], '/search')
    handler_kwargs = dict(work_dir=nb_app.notebook_dir)
    web_app.add_handlers(host_pattern, [
        (route_pattern, SearchHandler, handler_kwargs)
    ])



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.
from IPython.core.magic import Magics, magics_class, line_magic
from IPython.display import display, Javascript
import json

@magics_class
class InjectMagic(Magics):
    def _build_injection_js(self, notebook):
        '''
        Creates a series of JS commands to inject code and markdown cells from
        the passed notebook structure into the existing notebook sans output.
        '''
        js = ['var i = IPython.notebook.get_selected_index();']
        i = 0
        for cell in notebook['cells']:
            if cell['cell_type'] == 'markdown':
                js.append("var cell = IPython.notebook.insert_cell_below('markdown', i+{});".format(i))
                escaped_source = json.dumps(cell['source'].strip()).strip('"')
                js.append('cell.set_text("{}");'.format(escaped_source))
                js.append('cell.rendered = false; cell.render();')
                i += 1
            elif cell['cell_type'] == 'code':
                js.append("var cell = IPython.notebook.insert_cell_below('code', i+{});".format(i))
                escaped_input = json.dumps(cell['source'].strip()).strip('"')
                js.append('cell.set_text("{}");'.format(escaped_input))
                i += 1
        js.append('var self = this; setTimeout(function() { self.clear_output(); }, 0);')
        return '\n'.join(js)

    @line_magic
    def inject(self, line):
        '''
        Injects the rich help notebook associated with the named object as
        cells in this notebook.

            import urth.cms
            import mywb.some_notebook as sn
            %inject sn.some_recipe
        '''
        line = line.strip(' ()')
        segs = line.split('.')
        obj = self.shell.user_module
        for seg in segs:
            try:
                obj = getattr(obj, seg)
            except AttributeError:
                return 'Could not locate ' + line
        if hasattr(obj, '__richdoc__'):
            js = self._build_injection_js(obj.__richdoc__)
        elif hasattr(obj, 'help') and hasattr(obj.help, '__richdoc__'):
            js = self._build_injection_js(obj.help.__richdoc__)
        else:
            return 'Injection of the requested object is not yet supported'
        
        display(Javascript(js))



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.
'''
Extended from http://nbviewer.ipython.org/github/ipython/ipython/blob/master/examples/notebooks/Importing%20Notebooks.ipynb.
'''
import io, os, sys, types, re, warnings, hashlib
import nbformat
from nbformat import v4 as nb_v4
from nbformat import reader, converter
from nbconvert.exporters.html import HTMLExporter
from traitlets.config import Config
from IPython.display import display, HTML
from IPython.core.interactiveshell import InteractiveShell

UNBOUND_HELP_DOCSTRING_TMPL = 'Run "{name}()" to see this example. Run "%inject {name}()" to inject the example into your notebook.'
BOUND_HELP_DOCSTRING_TMPL = 'Run this function to see a rich example.'
API_DOCSTRING_TMPL = 'Run {name}.help() for a detailed example.'
MODULE_API_DOCSTRING_TMPL = 'Run help() for a detailed example.'
MODULE_DOCSTRING_TMPL = 'Run this function to see rich help about this importable notebook.'

export_html = HTMLExporter(config=Config({
    'CSSHTMLHeaderPreprocessor': {
        'enabled': False
    },
    'HTMLExporter': {
        'template_file': 'basic'
    }
}))


def rich_help():
    def _rich_help(*args):
        with warnings.catch_warnings():
            # ignore warnings about pandoc, nodejs, etc.
            warnings.filterwarnings('ignore', message='Node.js')
            output, resources = export_html.from_notebook_node(_rich_help.__richdoc__)
        # include style in every output; tried it global on import, but if the import is
        # re-run it's lost because there is no output on a repeat module import
        output = '''<style>
    .output .input_prompt,
    .output .output .output_prompt,
    .output .output .prompt { display: none; }
    .output .input_area pre { padding: 0.4em; }
</style>
''' + output
        display(HTML(output))
    return _rich_help

def convert_notebook(notebook):
    '''Converts IPython notebook to current version.'''
    return converter.convert(notebook, nbformat.current_nbformat)

class NotebookLoader(object):
    '''Module loader for IPython Notebooks with support for rich help.'''
    API_REX = re.compile('#\s*<api>')
    # help annotation acceptable at the start of the cell or on the line
    # immediately below a cell magic
    HELP_REX = re.compile('(%%[^\n]+\n)?(#\s*<help(:([^>]+))?>\s*)')

    def __init__(self, path, nb_path, inject_locals={}):
        self.shell = InteractiveShell.instance()
        self.path = path
        self.nb_path = nb_path
        self.inject_locals = inject_locals

    def create_rich_help_func(self):
        '''
        Build a help renderer with an attached __richdoc__ notebook.
        '''
        nb = nb_v4.new_notebook()
        f = rich_help()
        f.__richdoc__ = nb
        return f

    def attach_richdoc(self, obj, cell, prev, name):
        '''
        Attach a help() function to the object that renders rich help. If obj
        is None, don't attach the help renderer, just return it. In either case,
        append the cell markup and the previous cell (if it's markdown) to the
        help renderer.
        '''
        if obj is None:
            # print 'synthesizing function'
            f = self.create_rich_help_func()
            f.__doc__ = UNBOUND_HELP_DOCSTRING_TMPL.format(name=name)
        elif not hasattr(obj, 'help') and not hasattr(obj, '__richdoc__'):
            # print 'attaching as help'
            f = self.create_rich_help_func()
            if isinstance(obj, type):
                # can't assign __doc__ to methods on class or class itself
                obj.help = classmethod(f)
            else:
                obj.help = f
                tmpl = BOUND_HELP_DOCSTRING_TMPL if name is not None else MODULE_DOCSTRING_TMPL
                obj.help.__doc__ = tmpl.format(name=name)

                tmpl = API_DOCSTRING_TMPL if name is not None else MODULE_API_DOCSTRING_TMPL
                api_doc = tmpl.format(name=name)
                if obj.__doc__ is None:
                    obj.__doc__ = api_doc
                elif not obj.__doc__.endswith(api_doc):
                    obj.__doc__ += '\n' + api_doc
        elif hasattr(obj, '__richdoc__'):
            # print 'appending to synthesized function'
            f = obj
        elif hasattr(obj.help, '__richdoc__'):
            # print 'appending to help'
            f = obj.help
        else:
            raise RuntimeError

        if prev and prev.cell_type == 'markdown':
            f.__richdoc__.cells.append(prev)
        # add this cell to the help notebook if it's non-empty
        if cell.source.strip():
            f.__richdoc__.cells.append(cell)

        return f

    def eval_notebook(self, nb, mod):
        '''
        Evaluate notebook cells.
        '''
        mod.__dict__.update(self.inject_locals)
        prev = None
        if not len(nb.cells):
            # notebook is new and has no cells
            return
        for cell in nb.cells:
            if not mod.__doc__ and cell.cell_type == 'markdown':
                mod.__doc__ = cell.source
            elif cell.cell_type == 'code' and len(cell.source):
                if self.API_REX.match(cell.source):
                    # transform the input to executable Python
                    code = self.shell.input_transformer_manager.transform_cell(cell.source)
                    # run the code in the module
                    exec(code, mod.__dict__)
                else:
                    match = self.HELP_REX.match(cell.source)
                    if match:
                        # strip the cell input of the help tag marker
                        cell.source = cell.source[:match.start(2)] + cell.source[match.end(2):]
                        name = match.group(4)
                        if name is not None:
                            name = name.strip()
                            try:
                                obj = mod.__dict__[name]
                            except KeyError:
                                # attach to a new named function on the module itself
                                obj = self.attach_richdoc(None, cell, prev, name)
                                mod.__dict__[name] = obj
                            else:
                                # attach to an existing object, assuming that it comes
                                # before the help in the notebook
                                self.attach_richdoc(obj, cell, prev, name)
                        else:
                            # attach to the module itself
                            self.attach_richdoc(mod, cell, prev, name)

            # store previous
            prev = cell

    def load_module(self, fullname):
        '''
        Creates a module containing the exposed API content of a notebook
        by evaluating those notebook cells.
        '''
        # parse the notebook json
        with io.open(self.nb_path, 'r', encoding='utf-8') as f:
            notebook = reader.read(f)
        # convert to current notebook version
        notebook = convert_notebook(notebook)

        # create the module
        mod = types.ModuleType(fullname)
        mod.__file__ = self.nb_path
        mod.__package__ = '.'.join(fullname.split('.')[:-1])
        mod.__loader__ = self
        sys.modules[fullname] = mod

        # extra work to ensure that magics that would affect the user_ns
        # actually affect the notebook module's ns
        save_user_ns = self.shell.user_ns
        self.shell.user_ns = mod.__dict__

        try:
            self.eval_notebook(notebook, mod)
        finally:
            self.shell.user_ns = save_user_ns
        return mod

    def load_module_by_path(self):
        # parse the notebook json
        with io.open(self.nb_path, 'r', encoding='utf-8') as f:
            notebook = reader.read(f)
        # convert to current notebook version
        notebook = convert_notebook(notebook)

        # create the module and generate a hash from the name
        fullname = hashlib.sha1().hexdigest()
        mod = types.ModuleType(fullname)
        mod.__file__ = self.nb_path
        mod.__package__ = None
        mod.__loader__ = self
        sys.modules[fullname] = mod

        # extra work to ensure that magics that would affect the user_ns
        # actually affect the notebook module's ns
        save_user_ns = self.shell.user_ns
        self.shell.user_ns = mod.__dict__

        try:
            self.eval_notebook(notebook, mod)
        finally:
            self.shell.user_ns = save_user_ns
        return mod

class BlankPackageLoader(object):
    '''
    Acts like the root of an empty Python package.
    '''
    def __init__(self, path, *args):
        self.path = path

    def load_module(self, fullname):
        mod = types.ModuleType(fullname)
        mod.__path__ = self.path
        mod.__package__ = fullname
        mod.__loader__ = self
        sys.modules[fullname] = mod
        return mod

class NotebookFinder(object):
    '''
    Treats IPython Notebooks as importable modules.
    '''
    def __init__(self, loader_cls):
        self.loader_cls = loader_cls

    def find_module(self, fullname, path=None):
        '''
        Find a notebook, given its fully qualified name and an optional path
        '''
        loader = None
        # print('*** finding fullname:', fullname, 'path:', path)
        if fullname.startswith('mywb.'):
            name = fullname.rsplit('.', 1)[-1]
            path = [''] if path is None or not len(path) else path
            fullpath = os.path.join(*path)

            nb_path = os.path.join(fullpath, name + ".ipynb")
            if os.path.isfile(nb_path):
                loader = self.loader_cls(path, nb_path)
        return loader


class NotebookPathFinder(object):
    '''
    Treats the resources root folder and subdirectories as blank modules.
    '''
    def __init__(self, root, loader_cls):
        self.root = root
        self.loader_cls = loader_cls

    def find_module(self, fullname, ns_path=None):
        if fullname == 'mywb' and ns_path is None:
            # bootstrap the top level fake dir for mywb
            loader = self.loader_cls([self.root])
        elif fullname.startswith('mywb.') and ns_path:
            segs = fullname.split('.')
            path = ns_path._path if hasattr(ns_path, '_path') else ns_path
            fullpath = os.path.join(*(path + segs[-1:]))

            if os.path.isdir(fullpath):
                loader = self.loader_cls([fullpath])
            else:
                loader = None
        else:
            loader = None

        return loader

_enabled = None


def enable(root,
           notebook_path_loader_cls=BlankPackageLoader,
           notebook_loader_cls=NotebookLoader):
    '''
    Adds finders to the sys.meta_path to load Jupyter Notebooks stored
    under the given root path.
    '''
    global _enabled
    if _enabled is None:
        nb_path_finder = NotebookPathFinder(root, notebook_path_loader_cls)
        nb_finder = NotebookFinder(notebook_loader_cls)
        sys.meta_path.append(nb_path_finder)
        sys.meta_path.append(nb_finder)
        _enabled = (nb_path_finder, nb_finder)
    else:
        raise RuntimeError('loader already enabled')


def disable():
    '''
    Removes notebook finders from the sys.meta_path.
    '''
    global _enabled
    if _enabled is not None:
        sys.meta_path.remove(_enabled[0])
        sys.meta_path.remove(_enabled[1])
        _enabled = None
    else:
        raise RuntimeError('loader not enabled')


def load_notebook(path, inject_locals={}):
    '''
    Loads a notebook as a module given its absolute path. Returns a new instance
    of the module every time it is called unlike normal imports which
    sys.modules caches based on name.
    '''
    loader = NotebookLoader([], path, inject_locals=inject_locals)
    return loader.load_module_by_path()
